(window.webpackJsonp=window.webpackJsonp||[]).push([[197],{1375:function(e,t,a){"use strict";a.r(t),a.d(t,"useDatasetsViewModel",(function(){return o}));var n=a(3),r=(a(35),a(100)),s=a(182),c=a(283),u=a(284),o=function useDatasetsViewModel(){var e=Object(s.c)(!1),t=Object(u.a)().state,a=Object(r.useResolve)(c.a);Object(s.e)((function(){o()}));var o=function(){var t=Object(n.a)(regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(t){for(;;)switch(t.prev=t.next){case 0:return e.value=!0,t.next=3,a.execute();case 3:e.value=!1;case 4:case"end":return t.stop()}}),_callee)})));return function loadDatasets(){return t.apply(this,arguments)}}();return{datasets:t,isLoadingDatasets:e}}}}]);