(window.webpackJsonp=window.webpackJsonp||[]).push([[193],{1383:function(e,t,a){"use strict";a.r(t),a.d(t,"useDatasetSettingViewModel",(function(){return d}));var n=a(3),r=(a(35),a(182)),s=a(100),c=a(39),o=a(286),i=a(206),u=a(285),d=function useDatasetSettingViewModel(){var e=Object(r.c)(!1),t=Object(r.g)(),a=Object(r.f)().value.params.id,d=Object(u.a)().state,f=Object(s.useResolve)(o.a),l=function handleError(e){var t="";switch(e){case i.b.ERROR_FETCHING_DATASET_INFO:t="Can't get dataset info for dataset_id: ".concat(a);break;case i.b.ERROR_FETCHING_WORKSPACE_INFO:t="Can't get workspace info for dataset_id: ".concat(a);break;default:t="There was an error on fetching dataset info and workspace info. Please try again"}var n={message:t,numberOfChars:t.length,type:"error"};c.a.dispatch("notify",n)},p=function(){var r=Object(n.a)(regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(n){for(;;)switch(n.prev=n.next){case 0:return n.prev=0,e.value=!0,n.next=4,f.execute(a);case 4:n.next=10;break;case 6:n.prev=6,n.t0=n.catch(0),l(n.t0.response),t.push("/");case 10:return n.prev=10,e.value=!1,n.finish(10);case 13:case"end":return n.stop()}}),_callee,null,[[0,6,10,13]])})));return function loadDataset(){return r.apply(this,arguments)}}();return Object(r.b)((function(){p()})),{dataset:d,datasetId:a,isLoadingDataset:e}}}}]);